
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#include <cplex_solver.h>

abstract_solver *new_cplex_solver() { return new cplex_solver(); }

int cplex_solver::init_solver(CUDFVersionedPackageList *all_versioned_packages, int other_vars) {
  int status;


  nb_packages = all_versioned_packages->size();
  nb_vars = nb_packages + other_vars;
  this->all_versioned_packages = all_versioned_packages;

  /* Initialize the CPLEX environment */
  env = CPXopenCPLEX (&status);

  //  CPXsetintparam(env, CPX_PARAM_THREADS, 2);
  CPXsetintparam(env, CPX_PARAM_CONFLICTDISPLAY, 2); // Display more informations about conflicts
  CPXsetintparam(env, CPX_PARAM_SCRIND, 1);

  if ( env == NULL ) {
    char  errmsg[1024];
    fprintf (stderr, "Could not open CPLEX environment.\n");
    CPXgeterrorstring (env, status, errmsg);
    fprintf (stderr, "%s", errmsg);
    exit(-1);
  }

  /* Turn on output to the screen */
  status = CPXsetintparam (env, CPX_PARAM_SCRIND, CPX_ON);
  if ( status ) {
      fprintf (stderr, "Failure to turn on screen indicator, error %d.\n", status);
      exit(-1);
  }
  
  
   /* Turn on data checking */
  
  status = CPXsetintparam (env, CPX_PARAM_DATACHECK, CPX_ON);
  if ( status ) {
    fprintf (stderr, "Failure to turn on data checking, error %d.\n", status);
    exit(-1);
  }
  
   /* Enhance EPGAP to handle big values correctly */
  
  status = CPXsetdblparam (env, CPX_PARAM_EPGAP, 0.0);
  if ( status ) {
    fprintf (stderr, "Failure to set EPGAP, error %d.\n", status);
    exit(-1);
  }
  
  
  /* Create the problem. */

  lp = CPXcreateprob (env, &status, "lpex1");
  
  /* A returned pointer of NULL may mean that not enough memory
     was available or there was some other problem.  In the case of
     failure, an error message will have been written to the error
     channel from inside CPLEX.  In this example, the setting of
     the parameter CPX_PARAM_SCRIND causes the error message to
     appear on stdout.  */
  
  if ( lp == NULL ) {
    fprintf (stderr, "Failed to create LP.\n");
    exit(-1);
  }

  coefficient = (double *)malloc(nb_vars*sizeof(double));
  index = (int *)malloc(nb_vars*sizeof(int));
  lb = (double *)malloc(nb_vars*sizeof(double));
  ub = (double *)malloc(nb_vars*sizeof(double));
  vartype = (char *)malloc(nb_vars*sizeof(char));
  varname = (char **)malloc(nb_vars*sizeof(char *));

  if ((coefficient == (double *)NULL) ||
      (lb  == (double *)NULL) ||
      (ub  == (double *)NULL) ||
      (vartype  == (char *)NULL) ||
      (varname  == (char **)NULL) ||
      (index  == (int *)NULL))
    return -1;
  else
    return 0;
}

int cplex_solver::writelp(char *filename) { return CPXwriteprob (env, lp, filename, NULL); }

int cplex_solver::solve() {
  int mipstat;
  double objval;
  bool first_try = true;
  
  for (int i = 0; i < nb_objectives; i++) {
    if (CPXmipopt (env, lp)) 
      return 0;
  
    mipstat = CPXgetstat(env, lp);
    if (mipstat == CPXMIP_OPTIMAL) {
      if (i < nb_objectives - 1) {
	int  status = CPXgetobjval (env, lp, &objval);
	if ( status ) {
	  fprintf (stderr,"No MIP objective value available.  Exiting...\n");
	  exit(-1);
	}
	{
	  int indices[1];
	  double values[1];
	  indices[0] = i;
	  printf(">>>>>>> Objval = %f (int)%d\n", objval, (int)objval);
	  first_try = true;
	  values[0] = objval; // + 1; // +epsilon car sinon Row 'c1' infeasible, all entries at implied bounds.
	  status = CPXchgrhs(env, lp, 1, indices, values); 
	  if ( status ) {
	    fprintf (stderr,"Cannot change rhs for row %d.  Exiting...\n", i);
	    exit(-1);
	  }
	}
	status = CPXchgobj(env, lp, nb_vars, obj_index, objs[i+1]);
	if ( status ) {
	  fprintf (stderr,"Cannot change objective value.  Exiting...\n");
	  exit(-1);
	}
	{
	  char buffer[1024];
	  sprintf(buffer, "cplexpbs%d.lp", i);
	  CPXwriteprob (env, lp, buffer, NULL);
	}
      } else
	return 1;
    } else if (0 && (mipstat == 103) && (i > 0) && (first_try)) {
      int indices[1];
      double values[1];
      i--;
      indices[0] = i;
      printf("(second attempt)>>>>>>> Objval = %f\n", objval+1);
      first_try = false;
      values[0] = objval + 1; // +epsilon car sinon Row 'c1' infeasible, all entries at implied bounds.
      status = CPXchgrhs(env, lp, 1, indices, values); 
      if ( status ) {
	fprintf (stderr,"Cannot change rhs for row %d.  Exiting...\n", i);
	exit(-1);
      }
    } else {
      fprintf(stderr, "CPLEX solution status = %d\n", mipstat);
      return 0;
    }
  }

  return 0;
}

double cplex_solver::objective_value() { 
  double objval;
  int  status = CPXgetobjval (env, lp, &objval);
  if ( status ) {
    fprintf (stderr,"No MIP objective value available.  Exiting...\n");
    exit(-1);
  }
  return objval; 
}

int cplex_solver::init_solutions() {
  int status;
  //  int cur_numrows = CPXgetnumrows (env, lp);
  int cur_numcols = CPXgetnumcols (env, lp);

   solution = (double *)malloc(nb_vars*sizeof(double));
   status = CPXgetx (env, lp, solution, 0, cur_numcols-1);
   if ( status ) {
      fprintf (stderr, "Failed to get optimal integer.\n");
      exit(-1);
   }
  return 0; 
}

double cplex_solver::get_solution(CUDFVersionedPackage *package) { return solution[package->rank]; }

int cplex_solver::begin_objective(void) { 
  int i = 0;
  for (CUDFVersionedPackageListIterator ipkg = all_versioned_packages->begin(); ipkg != all_versioned_packages->end(); ipkg++) {
    coefficient[i] = 0;
    lb[i] = 0; 
    ub[i] = 1;
    vartype[i] = 'B';
    varname[(*ipkg)->rank] = (*ipkg)->versioned_name;
    i++;
  }
  for (i = nb_packages; i < nb_vars; i++) {
    char *name;
    char buffer[20];

    sprintf(buffer, "x%d", i);
    if ((name = (char *)malloc(strlen(buffer)+1)) == (char *)NULL) {
      fprintf(stderr, "CUDF error: can not alloc memory for variable name in glpk_solver::end_objective.\n");
      exit(-1);
    }
    strcpy(name, buffer);

    coefficient[i] = 0;
    lb[i] = 0; 
    ub[i] = 1;
    vartype[i] = 'B';
    varname[i] = name;
  }

  CPXchgobjsen (env, lp, CPX_MIN);  // Problem is minimization
  return 0; 
}

double cplex_solver::get_obj_coeff(CUDFVersionedPackage *package) { return coefficient[package->rank]; }

int cplex_solver::set_obj_coeff(CUDFVersionedPackage *package, double value) { coefficient[package->rank] = value; return 0; }

int cplex_solver::set_obj_coeff(int rank, double value) { coefficient[rank] = value; return 0; };

int cplex_solver::end_objective(void) {
  // Build the objective function
  nb_objectives++;
  status = CPXnewcols (env, lp, nb_vars, coefficient, lb, ub, vartype, varname);
  if (status) {
    fprintf(stderr, "generate_milp: cannot create objective function.\n");
    exit(-1);
  }
  return 0;
}

int cplex_solver::new_objective(void) {

  if (obj_index == (int *)NULL) {
    obj_index = (int *)malloc(nb_vars*sizeof(int));
    if (obj_index == (int *)NULL) return -1;
    for (int i = 0; i < nb_vars; i++) obj_index[i] = i;
  }

  if (nb_objectives == 1) {
    double ub = 0.0;

    for (int i = 0; i < nb_vars; i++) if (coefficient[i] > 0.0) ub += coefficient[i];

    begin[0] = 0;
    begin[1] = nb_vars - 1;
    rhs[0] = ub;
    status = CPXaddrows(env, lp, 0, 1, nb_vars, rhs, "L", begin, obj_index, coefficient, NULL, NULL);
    if (status) { fprintf(stderr, "generate_milp: cannot create constraint.\n"); exit(-1); }

    objs[nb_objectives-1] = (double *)malloc(nb_vars*sizeof(double));
    if (objs[nb_objectives - 1] == (double *)NULL) return -1;
    memcpy(objs[nb_objectives-1], coefficient, nb_vars*sizeof(double));
  }

  for (int i = 0; i < nb_vars; i++) coefficient[i] = 0.0;

  return 0;
}

int cplex_solver::add_objective(void) {
  double ub = 0.0;

  for (int i = 0; i < nb_vars; i++) if (coefficient[i] > 0.0) ub += coefficient[i];

  begin[0] = 0;
  begin[1] = nb_vars - 1;
  rhs[0] = ub;
  status = CPXaddrows(env, lp, 0, 1, nb_vars, rhs, "L", begin, obj_index, coefficient, NULL, NULL);
  if (status) { fprintf(stderr, "generate_milp: cannot create constraint.\n"); exit(-1); }

  objs[nb_objectives] = (double *)malloc(nb_vars*sizeof(double));
  if (objs[nb_objectives] == (double *)NULL) return -1;
  memcpy(objs[nb_objectives], coefficient, nb_vars*sizeof(double));
  nb_objectives++;

  return 0;
}

int cplex_solver::begin_add_constraints(void) { begin[0] = 0; return 0; }

int cplex_solver::new_constraint(void) { rank = 0; return 0; }

double cplex_solver::get_constraint_coeff(CUDFVersionedPackage *package) { 
  for(int i =0; i < rank; i++)
    if (index[i] == package->rank) return coefficient[i];
  return 0;
}

int cplex_solver::set_constraint_coeff(CUDFVersionedPackage *package, double value) { 
  coefficient[rank] = value;
  index[rank] = package->rank;
  rank++;
  return 0;
}

int cplex_solver::set_constraint_coeff(int therank, double value) { 
  coefficient[rank] = value;
  index[rank] = therank;
  rank++;
  return 0;
}

int cplex_solver::add_constraint_geq(double bound) {
  if (rank > 0) {
    begin[1] = rank;
    rhs[0] = bound;
    status = CPXaddrows(env, lp, 0, 1, rank, rhs, "G", begin, index, coefficient, NULL, NULL);
    if (status) { fprintf(stderr, "generate_milp: cannot create constraint.\n"); exit(-1); }
  }
  return 0;
}

int cplex_solver::add_constraint_leq(double bound) {
  if (rank > 0) {
    begin[1] = rank;
    rhs[0] = bound;
    status = CPXaddrows(env, lp, 0, 1, rank, rhs, "L", begin, index, coefficient, NULL, NULL);
    if (status) { fprintf(stderr, "generate_milp: cannot create constraint.\n"); exit(-1); }
  }
  return 0;
}

int cplex_solver::add_constraint_eq(double bound) {
  if (rank > 0) {
    begin[1] = rank;
    rhs[0] = bound;
    status = CPXaddrows(env, lp, 0, 1, rank, rhs, "E", begin, index, coefficient, NULL, NULL);
    if (status) { fprintf(stderr, "generate_milp: cannot create constraint.\n"); exit(-1); }
  }
  return 0;
}

int cplex_solver::end_add_constraints(void) { CPXwriteprob (env, lp, "cplexpbs.lp", NULL); return 0; }
