
/*******************************************************/
/* (c) Claude Michel I3S (UNSA-CNRS) 2009,2010         */
/* Distributed under GPLv3                             */
/*******************************************************/


#include <cudf.h>
#include <abstract_solver.h>
#include <constraint_generation.h>

extern abstract_solver *new_pblib_solver(char *pbsolver);
extern abstract_solver *new_ampl_solver(char *amplsolver);
#ifdef USECPLEX 
extern abstract_solver *new_cplex_solver();
#endif
#ifdef USELPSOLVE 
extern abstract_solver *new_lpsolve_solver();
#endif
#ifdef USEGLPK
extern abstract_solver *new_glpk_solver(bool use_exact);
#endif
#ifdef USECPO
extern abstract_solver *new_cpo_solver();
#endif

bool silent = true;

void print_help() {
  fprintf(stderr, "cudfsolver [(-i <input_file>)*][-o <ouput_file>][-fo]["
#ifdef USECPLEX
	  "-cplex|"
#endif
#ifdef USELPSOLVE
	  "-lpsolve|"
#endif
#ifdef USEGLPK
	  "-glpk|-glpkQ|"
#endif
#ifdef USECPLEX
	  "-cpo|"
#endif
	  "-pblib <pbsolver>|-ampl <amplsolver>|-nosolve]"
	  "[-v][-conservative|-fresher|-sfresher|-ffresher|-paranoid|-trendy][-h]\n");
  fprintf(stderr, "This software is distributed under GNU GPLv3 and was\n"
	  "partially supported by the European Community's 7th Framework Programme (FP7/2007-2013),\n" 
	  "MANCOOSI project, grant agreement n. 214898.\n");
  fprintf(stderr, "Usual call: cudfsolver -i <input_file> -o outputfile <criteria option> <solver option> <other options>?\n");
  fprintf(stderr, "file options:\n");
  fprintf(stderr, " -i <input_file>: set the input file to <input_file> (by default attempt to read on stdin)\n");
  fprintf(stderr, " -o <output_file>: set the output file to <output_file> (by default write on stdout)\n");
  fprintf(stderr, "solver options:\n");
#ifdef USECPLEX
  fprintf(stderr, " -cplex: use IBM ILOG cplex solver\n");
#endif
#ifdef USELPSOLVE
  fprintf(stderr, " -lpsolve: use lpsolve solver\n");
#endif
#ifdef USEGLPK
  fprintf(stderr, " -glpk: use glpk solver\n");
  fprintf(stderr, " -glpkQ: use glpk rational solver\n");
#endif
#ifdef USECPO
  fprintf(stderr, " -cpo: use cpo IBM ILOG solver\n");
#endif
  fprintf(stderr, " -pblib <pbsolver>: use pseudo boolean solver <pbsolver>\n");
  fprintf(stderr, " -ampl <amplsolver>: use ampl solver <amplsolver>\n");
  fprintf(stderr, " -nosolve: do not solve the problem (for debug purpose)\n");
  fprintf(stderr, "criteria options:\n");
  fprintf(stderr, " -paranoid: paranoid competition criteria (aggregate version)\n");
  fprintf(stderr, " -trendy: trendy competition criteria (aggregate version)\n");
#ifdef USECPLEX
  fprintf(stderr, " -lexparanoid: paranoid competition criteria (fully lexicographic version (require cplex))\n");
  fprintf(stderr, " -lextrendy: trendy competition criteria (fully lexicographic version (require cplex))\n");
#endif
  fprintf(stderr, " -conservative|-fresher|-sfresher|-ffresher: undocumented criteria\n");
  fprintf(stderr, "other options:\n");
  fprintf(stderr, " -fo: full solution output\n");
  fprintf(stderr, " -v: verbose mode\n");
  fprintf(stderr, " -h: print this help\n");

}

int main(int argc, char *argv[]) {
  FILE *output_file = (FILE *)NULL;
  FILE *output_installed = (FILE *)NULL;
  FILE *output_removed = (FILE *)NULL;
  abstract_solver *solver = (abstract_solver *)NULL;
  int criteria = CRIT_DEFAULT;
  bool nosolve = false;
  bool got_input = false;
  bool fulloutput = false;

  if (argc > 1) 
    for (int i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-i") == 0) {
	i++;
	if (i < argc) {
	  FILE *input_file = (FILE *)NULL;

	  if ((input_file = fopen(argv[i], "r")) == (FILE *)NULL) {
	    fprintf(stderr, "Cannot open file %s as input file.\n", argv[i]);
	    exit(-1);
	  } else {
	    got_input = true;
	    switch (parse_cudf(input_file)) {
	    case 0: break;
	    case 1: fprintf(stderr, "Invalid input in problem.\n"); exit(-1);
	    case 2: fprintf(stderr, "Parser memory issue.\n"); exit(-1);
	    }
	    fclose(input_file);
	  }
	}
      } else if (strcmp(argv[i], "-o") == 0) {
	i++;
	if (i < argc) {
	  if ((output_file = fopen(argv[i], "w")) == (FILE *)NULL) {
	    fprintf(stderr, "Cannot open file %s as ouput file.\n", argv[i]);
	    exit(-1);
	  }
	}
      } else if (strcmp(argv[i], "-fo") == 0) {
	fulloutput = true;
      } else if (strcmp(argv[i], "-v") == 0) {
	silent = false;
      } else if (strcmp(argv[i], "-conservative") == 0) {
	criteria = CRIT_CONSERVATIVE;
      } else if (strcmp(argv[i], "-fresher") == 0) {
	criteria = CRIT_FRESHER;
      } else if (strcmp(argv[i], "-sfresher") == 0) {
	criteria = CRIT_SFRESHER;
      } else if (strcmp(argv[i], "-ffresher") == 0) {
	criteria = CRIT_FFRESHER;
      } else if (strcmp(argv[i], "-paranoid") == 0) {
	criteria = CRIT_PARANOID;
      } else if (strcmp(argv[i], "-trendy") == 0) {
	criteria = CRIT_TRENDY;
      } else if (strcmp(argv[i], "-trendy2") == 0) {
	criteria = CRIT_TRENDY2;
      } else if (strcmp(argv[i], "-lexparanoid") == 0) {
	criteria = CRIT_LEXPARANOID;
      } else if (strcmp(argv[i], "-lextrendy") == 0) {
	criteria = CRIT_LEXTRENDY;
      } else if (strcmp(argv[i], "-lextrendy2") == 0) {
	criteria = CRIT_LEXTRENDY2;
      } else if (strcmp(argv[i], "-nosolve") == 0) {
	nosolve = true;
      } else if (strcmp(argv[i], "-h") == 0) {
	print_help();
	exit(-1);
      } else if (strcmp(argv[i], "-pblib") == 0) {
	if (++i < argc) {
	  solver = new_pblib_solver(argv[i]);
	} else {
	  fprintf(stderr, "-pblib option require a PB solver: -pblib <pbsolver>\n");
	  exit(-1);
	}
      } else if (strcmp(argv[i], "-ampl") == 0) {	
        if (++i < argc) {
	  solver = new_ampl_solver(argv[i]);
	} else {
	  fprintf(stderr, "-ampl option require a PB solver: -ampl <amplsolver>\n");
	  exit(-1);
	}
#ifdef USECPLEX
      } else if (strcmp(argv[i], "-cplex") == 0) {
	solver = new_cplex_solver();
#endif
#ifdef USECPO
      } else if (strcmp(argv[i], "-cpo") == 0) {
	solver = new_cpo_solver();
#endif
#ifdef USELPSOLVE
      } else if (strcmp(argv[i], "-lpsolve") == 0) {
	solver = new_lpsolve_solver();
#endif
#ifdef USEGLPK
      } else if (strcmp(argv[i], "-glpk") == 0) {
	solver = new_glpk_solver(false);
#endif
#ifdef USEGLPK
      } else if (strcmp(argv[i], "-glpkQ") == 0) {
	solver = new_glpk_solver(true);
#endif
      } else {
	fprintf(stderr, "Unrecognized option %s\n", argv[i]);
	print_help();
	exit(-1);
      }
    }

  //  if (input_file == (FILE *)NULL) input_file = stdin;
  if (output_file == (FILE *)NULL) output_file = stdout;

  if (! got_input) {
    switch (parse_cudf(stdin)) {
    case 0: break;
    case 1: fprintf(stderr, "Invalid input in problem.\n"); exit(-1);
    case 2: fprintf(stderr, "Parser memory issue.\n"); exit(-1);
    }
  }

  if (! silent) {
    fprintf(stdout, "================================================================\n");
    if (properties.size() > 0) {
      fprintf(stdout, "preamble:\n");
      print_properties(stdout, &properties);
      fprintf(stdout, "\n");
    }
    fprintf(stdout, "# %d versioned packages:\n\n", (int)all_packages.size());
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
      print_versioned_package(stdout, *ipkg, false);
    print_problem(stdout, the_problem);

    fprintf(stdout, "================================================================\n");
    fprintf(stdout, "%d virtual packages:\n\n", (int)all_virtual_packages.size());
    for (CUDFVirtualPackageListIterator vpkg = all_virtual_packages.begin(); vpkg != all_virtual_packages.end(); vpkg++)
      print_virtual_package(stdout, *vpkg);
    fprintf(stdout, "================================================================\n");

    /*
    fprintf(stdout, "================================================================\n");
    fprintf(stdout, "# remove pbs:\n\n");
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
      if ((*ipkg)->installed) printf("remove: %s\n", (*ipkg)->name); ;
    fprintf(stdout, "================================================================\n");
    fprintf(stdout, "# update pbs:\n\n");
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
      if ((*ipkg)->installed)
	if ((*ipkg)->virtual_package->all_versions.size() >= 2) printf("update: %s\n", (*ipkg)->name); ;
    fprintf(stdout, "================================================================\n");
    */
  }

  if (solver == (abstract_solver *)NULL) 
#ifdef USECPLEX
    solver = new_cplex_solver();
#else
  #ifdef USEGLPK
    solver = new_glpk_solver(false);
  #else
    #ifdef USELPSOLVE
    solver = new_lpsolve_solver();
    #else
    { fprintf(stderr, "Error: No solver defined\n");  print_help(); exit(-1); }
    #endif
  #endif
#endif

  if ((generate_constraints(*solver, criteria) == 0) && (! nosolve) && (solver->solve())) {

    solver->init_solutions();

    double obj = solver->objective_value();
    if (! silent) {
      fprintf(stdout, "================================================================\n");
      printf("Objective value: %f\n", obj);

      for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
	printf("%s = %f\n", (*ipkg)->versioned_name, solver->get_solution(*ipkg));
      
      fprintf(stdout, "================================================================\n");
    
      fprintf(output_file, "\n");
    }

    if (! silent) {
      output_installed = fopen("installed.txt", "w");
      output_removed = fopen("removed.txt", "w");
    }

    if (properties.size() > 0) {
      fprintf(output_file, "\npreamble: \n");
      print_properties(output_file, &properties);
      fprintf(output_file, "\n\n");
    }

    int nb_installed = 0;
    int nb_removed = 0;
    int nb_newinstalled = 0;
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
      if (solver->get_solution(*ipkg) > 0.7) { 
	nb_installed++; 
	if (! (*ipkg)->installed) { 
	  nb_newinstalled++; 
	  if (output_installed != (FILE *)NULL) print_versioned_package_as_installed(output_installed, (*ipkg), true); 
	}
	print_versioned_package_as_installed(output_file, (*ipkg), true); 
      } else {
	if (fulloutput) print_versioned_package_with_install(output_file, (*ipkg), 0, true);
	if ((*ipkg)->installed) { 
	  nb_removed++; 
	  if (output_removed != (FILE *)NULL) print_versioned_package(output_removed, (*ipkg), true); 
	}
      }
      
    //    if (fulloutput) { print_problem(output_file, the_problem); fprintf(output_file, "\n"); }

    if (! silent) {
      fclose(output_installed);
      fclose(output_removed);
    }

    fprintf(output_file, "# problem = %zu packages, %zu installed packages.\n", all_packages.size(), installed_packages.size());
    fprintf(output_file, "# request = %zu install, %zu remove, %zu upgrade.\n", 
	    (the_problem->install == (CUDFVpkgList *)NULL)?0:the_problem->install->size(), 
	    (the_problem->remove  == (CUDFVpkgList *)NULL)?0:the_problem->remove->size(), 
	    (the_problem->upgrade == (CUDFVpkgList *)NULL)?0:the_problem->upgrade->size());
    fprintf(output_file, "# solution = %d installed packages, %d new installed, %d has been removed.\n", nb_installed, nb_newinstalled, nb_removed);
    fprintf(output_file, "# objective value = %f.\n", obj);

    /*
    fprintf(stdout, "================================================================\n");
    fprintf(stdout, "# remove pbs:\n\n");
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
      if (solver->get_solution(*ipkg) > 0.7) printf("remove: %s\n", (*ipkg)->name); ;
    fprintf(stdout, "================================================================\n");
    fprintf(stdout, "# update pbs:\n\n");
    for (CUDFVersionedPackageListIterator ipkg = all_packages.begin(); ipkg != all_packages.end(); ipkg++)
      if (solver->get_solution(*ipkg) > 0.7)
	if ((*ipkg)->virtual_package->all_versions.size() >= 2) printf("update: %s\n", (*ipkg)->name); ;
    fprintf(stdout, "================================================================\n");
    */
  } else {
    fprintf(stdout, "================================================================\n");
    fprintf(stdout, "No solution found.\n");
    fprintf(output_file, "FAIL\n");
    fprintf(output_file, "No solution found.\n");
  }
  exit(0);
}
